from collections.abc import Iterator
from dataclasses import dataclass

import ida_funcs
import ida_segment
from ida_funcs import func_t
from ida_segment import segment_t


@dataclass
class Segment:
    name: str
    start_ea: int
    end_ea: int
    size: int
    cls: str

    def __repr__(self):
        return f"[{self.start_ea:#x}-{self.end_ea:#x}] {self.name} size:{self.size} cls:{self.cls}"

    def __hash__(self) -> int:
        return hash((self.start_ea, self.end_ea))

    @staticmethod
    def from_segment(segment: segment_t) -> "Segment":
        return Segment(
            ida_segment.get_segm_name(segment),
            segment.start_ea,
            segment.end_ea,
            segment.end_ea - segment.start_ea,
            ida_segment.get_segm_class(segment),
        )

    def functions(self) -> Iterator[func_t]:
        """Iterate all functions that are defined in the given segment"""
        func_ea = self.start_ea
        func = ida_funcs.get_func(func_ea)
        if func is None:
            # No function there, try to get the next one
            func = ida_funcs.get_next_func(func_ea)

        while func is not None and func.start_ea < self.end_ea:
            yield func
            func = ida_funcs.get_next_func(func.start_ea)

    @property
    def base_name(self) -> str:
        """If the name is a.b.c:__d then return a.b.c ; Otherwise, return the original name"""
        return self.name.split(":")[0]


def get_segments(cls: str | None = None) -> list[Segment]:
    """Get all segments in the executable, optionally filter it by specific class"""
    segments = []
    for i in range(ida_segment.get_segm_qty()):
        seg = Segment.from_segment(ida_segment.getnseg(i))
        if cls is None or seg.cls == cls:
            segments.append(seg)
    return segments


def get_segment_by_name(name: str) -> Segment | None:
    """Get a segment with the given name or none if non-existent"""
    seg = ida_segment.get_segm_by_name(name)
    if seg is None:
        return None
    return Segment.from_segment(seg)


def get_segment_by_ea(ea: int) -> Segment | None:
    """Get a segment that contains the given EA or None if it does not exist"""
    seg = ida_segment.getseg(ea)
    if seg is None:
        return None
    return Segment.from_segment(seg)



import ida_bytes
from ida_hexrays import ITP_COLON, ITP_SEMI, cfunc_t, treeloc_t


def get_asm_comment(ea: int, repeatable: bool = False) -> str | None:
    """Get the comment of the given ea or none if it has no comment"""
    return ida_bytes.get_cmt(ea, repeatable)


def set_asm_comment(ea: int, comment: str, repeatable: bool = False) -> bool:
    """Set the comment for the given ea"""
    return ida_bytes.set_cmt(ea, comment, repeatable)


def set_psuedocode_comment(address: int, cfunc: cfunc_t, comment: str) -> bool:
    """Try to set comment in psuedocode"""
    eamap = cfunc.get_eamap()
    obj_addr = eamap[address][0].ea

    # get a ctree location object to place a comment there
    tl = treeloc_t()
    tl.ea = obj_addr

    # since the public documentation on IDAs APIs is crap and I don't know any other way, we have to brute force the item preciser
    # we do this by setting the comments with different idaapi.ITP_* types until our comment does not create an orphaned comment
    for itp in range(ITP_SEMI, ITP_COLON):
        tl.itp = itp
        cfunc.set_user_cmt(tl, comment)
        cfunc.save_user_cmts()
        # # apparently you have to cast cfunc to a string, to make it update itself
        cfunc.__str__()
        if not cfunc.has_orphan_cmts():
            cfunc.save_user_cmts()
            return True
        cfunc.del_orphan_cmts()
    return False



from ida_hexrays import (
    mcallarg_t,
    mop_t,
)
from ida_typeinf import tinfo_t


def from_mop(mop: mop_t, arg_type: tinfo_t) -> mcallarg_t:
    """Given existing `mop`, wrap it in a `callarg` of type `arg_type`"""
    arg = mcallarg_t()
    arg.copy_mop(mop)
    arg.type = arg_type
    return arg



from .. import PacClient
from .definitions import ExplorerProtocol, VtableXrefTuple, get_explorer_instance


class PacXplorerNGPacClient(PacClient):
    plugin_name = "pacxplorerng"

    def __init__(self):
        self.ensure_pac_plugin_installed()
        self.explorer: ExplorerProtocol = get_explorer_instance()
        self.explorer.analyze(True)

    def _pac_candidates_from_movk(self, movk_ea: int) -> list[VtableXrefTuple]:
        return self.explorer.movk_to_functions(movk_ea)

    def pac_xrefs_to_func(self, func_ea: int) -> list[int]:
        return self.explorer.function_to_movks(func_ea)



import ida_hexrays
import ida_kernwin
import ida_moves
import idaapi
from ida_hexrays import cexpr_t, cfunc_t
from ida_kernwin import Choose, place_t, simpleline_place_t


def refresh_pseudocode_widgets() -> None:
    """Refresh all pseudocode widgets in IDA Pro, forcing redecompiling."""
    for name in "ABCDEFGHIJKLMNOPQRSTUVWXY":
        widget = idaapi.find_widget(f"Pseudocode-{name}")
        if widget is not None:
            refresh_widget(widget)


def refresh_widget(widget: "TWidget *") -> None:  # noqa: F722
    """Refresh a given widget."""
    vdui: idaapi.vdui_t = idaapi.get_widget_vdui(widget)
    if vdui is None:
        return
    vdui.refresh_view(True)


def get_current_citem() -> cexpr_t | None:
    """Get the current citem in the active pseudocode window."""
    w = ida_kernwin.get_current_widget()
    if ida_kernwin.get_widget_type(w) != ida_kernwin.BWN_PSEUDOCODE:
        return None

    vu = ida_hexrays.get_widget_vdui(w)
    if vu is None:
        return None

    if not vu.get_current_item(ida_hexrays.USE_KEYBOARD) or not vu.item.is_citem():
        return None

    return vu.item.e


def get_current_function() -> cfunc_t | None:
    """Get the current function in the active pseudocode window."""
    w = ida_kernwin.get_current_widget()
    if ida_kernwin.get_widget_type(w) != ida_kernwin.BWN_PSEUDOCODE:
        return None

    vu = ida_hexrays.get_widget_vdui(w)
    if vu is None:
        return None

    return vu.cfunc


def jump_to(ea: int) -> None:
    """Jump to a given address in the current view."""
    ida_kernwin.jumpto(ea)


def jump_to_coords(widget, line: int, col: int) -> bool:
    """In the given widget, jump to the given (line, col)"""
    e = ida_moves.lochist_entry_t()
    if not ida_kernwin.get_custom_viewer_location(e, widget):
        print("Failed to get view location")
        return False
    place: place_t = e.place()
    if place is None:
        print("Failed to get place from lochist")
        return False

    print(place.lnnum)

    sl_place: simpleline_place_t = place_t.as_simpleline_place_t(place)
    if sl_place is None:
        print("Failed to get simple line place")
        return False

    sl_place.n = line
    e.set_place(sl_place)
    return ida_kernwin.jumpto(widget, sl_place, col, line)


def jump_to_coords_in_current_widget(line: int, col: int) -> bool:
    """jump to the given (line, col) in the current widget"""
    widget = ida_kernwin.get_current_widget()
    if widget is None:
        print("Failed to find current windet")
        return False
    return jump_to_coords(widget, line, col)


def show_yes_no_dialog(prompt: str) -> bool:
    """Show yes no dialog for the given {prompt}. Return whether yes button was selected"""
    return ida_kernwin.ask_yn(ida_kernwin.ASKBTN_NO, prompt) == ida_kernwin.ASKBTN_YES


def show_string_input(prompt: str, default: str = "") -> str | None:
    """Show a prompt for string input, returning user result"""
    return ida_kernwin.ask_str(default, ida_kernwin.HIST_CMT, prompt)


class EAChoose(Choose):
    """A chooser for data of the type <ea>:<description>"""

    def __init__(
        self,
        title: str,
        items: list[tuple[int, str]],
        col_names: tuple[str, str] = ("Address", "Name"),
        flags: int = 0,
        modal=False,
        embedded: bool = False,
        width: int | None = None,
        height: int | None = None,
    ):
        Choose.__init__(
            self,
            title,
            [[col_names[0], 10 | Choose.CHCOL_EA], [col_names[1], 40 | Choose.CHCOL_FNAME]],
            flags=flags | Choose.CH_RESTORE,
            embedded=embedded,
            width=width,
            height=height,
        )
        self.items = items
        self.modal = modal

    def OnInit(self):
        return True

    def OnGetSize(self):
        return len(self.items)

    def OnGetLine(self, n):
        ea, name = self.items[n]
        return hex(ea), name

    def OnGetEA(self, n):
        return self.items[n][0]

    def OnSelectLine(self, n):
        ea = int(self.items[n][0])
        ida_kernwin.jumpto(ea)
        return (Choose.NOTHING_CHANGED,)

    def show(self):
        ok = self.Show(self.modal) >= 0
        return ok



import idaapi

from idahelper import segments


def is_kernelcache() -> bool:
    """Check if the current file is a kernel cache"""
    file_type = idaapi.get_file_type_name()
    return "kernelcache" in file_type and "ARM64" in file_type


def is_objc() -> bool:
    """Check if the current file is an Objective-C binary"""
    return any(seg.name.startswith("__objc") or "__objc" in seg.name for seg in segments.get_segments())


def is_arm64() -> bool:
    """Check if the current file is an ARM64 binary"""
    file_type = idaapi.get_file_type_name()
    return "ARM64" in file_type



from collections.abc import Iterable, Iterator
from dataclasses import dataclass

import ida_funcs
import idc
from ida_typeinf import tinfo_t

from idahelper import functions, memory, tif


def demangle(symbol: str, strict: bool = False) -> str | None:
    """Demangle cpp symbol."""
    res = idc.demangle_name(symbol, idc.get_inf_attr(idc.INF_SHORT_DEMNAMES))
    if strict:
        return res
    return res or symbol


def demangle_name_only(symbol: str, strict: bool = False) -> str | None:
    """Demangle cpp symbol, return name of the function (including class)"""
    res = demangle(symbol, strict)
    if res is not None:
        return res.split("(")[0]
    return None


def demangle_class_only(symbol: str, strict: bool = False) -> str | None:
    """Demangle cpp symbol, return name of the class"""
    class_and_name = demangle_class_and_name(symbol, strict)
    if class_and_name is None:
        return None
    return class_and_name[0]


def demangle_class_and_name(symbol: str, strict: bool = False) -> tuple[str, str] | None:
    """Demangle cpp symbol, return tuple of class name and method"""
    name = demangle_name_only(symbol, strict)
    if name is None:
        return None
    # Expected Class::methodName or Class::innerClass::methodName
    last_double_colon = name.rfind("::")
    if last_double_colon == -1:
        return None
    return name[:last_double_colon], name[last_double_colon + 2 :]


def vtable_location_from_type(cpp_type: tinfo_t) -> int | None:
    """Find the location of the "\\`vtable'TYPE" for the given `cpp_type`"""
    # noinspection PyTypeChecker
    type_name: str = cpp_type.get_type_name()  # pyright: ignore[reportAssignmentType]
    return memory.ea_from_name(f"__ZTV{len(type_name)}{type_name}")


def class_name_from_vtable_name(symbol: str) -> str | None:
    """Given the name of the vtable symbol, return the class name"""
    vtable_demangled_name = demangle(symbol)
    if vtable_demangled_name and vtable_demangled_name.startswith("`vtable for'"):
        return vtable_demangled_name[12:]
    else:
        return None


def type_from_vtable_name(symbol: str) -> tinfo_t | None:
    """Given the name of the vtable symbol, return the cpp type"""
    cls_name = class_name_from_vtable_name(symbol)
    return tif.from_struct_name(cls_name) if cls_name else None


def iterate_vtables(skip_metaclass: bool = True) -> Iterator[tuple[str, int]]:
    """Iterate over all vtables in the database, yielding class name and vtable ea."""
    for ea, name in memory.names():
        cls = class_name_from_vtable_name(name)
        if cls is not None and (skip_metaclass or not cls.endswith("::MetaClass")):
            yield cls, ea


def get_all_cpp_classes() -> list[tuple[tinfo_t, int]]:
    """Return map between cpp type to its vtable ea"""
    d: list[tuple[tinfo_t, int]] = []
    for class_name, vtable_ea in iterate_vtables():
        cls = tif.from_struct_name(class_name)
        if cls is not None:
            d.append((cls, vtable_ea))
    return d


@dataclass
class VTableItem:
    index: int
    vtable_offset: int
    func_ea: int
    func_name: str
    demangled_func_name: str


def iterate_vtable(vtable_ea: int, skip_reserved: bool = True, raise_on_error: bool = False) -> Iterable[VTableItem]:
    """Iterate over the vtable at `vtable_ea`, yielding VTableItem for each function."""
    if memory.qword_from_ea(vtable_ea) != 0:
        error_msg = f"Expected null on offset 0, received: {memory.qword_from_ea(vtable_ea):X}"
        if raise_on_error:
            raise MemoryError()
        return

    current_ea = vtable_ea + 2 * memory.PTR_SIZE
    i = 0
    while (func_addr := memory.qword_from_ea(current_ea)) != 0:
        if ida_funcs.get_func(func_addr):
            mangled_name = memory.name_from_ea(func_addr)
        elif not (mangled_name := memory.name_from_imported_ea(func_addr)):
            error_msg = f"{vtable_ea:X}: Failed to get func from vtable at {current_ea:X}. Data: {func_addr:X}"
            if raise_on_error:
                raise MemoryError(error_msg)
            print(f"[Error] {error_msg}")
            return

        demangled_func_name = demangle_name_only(mangled_name, strict=False) or ""
        if not skip_reserved or "::_RESERVED" not in demangled_func_name:
            yield VTableItem(
                index=i,
                vtable_offset=current_ea - vtable_ea,
                func_ea=func_addr,
                func_name=mangled_name,
                demangled_func_name=demangled_func_name,
            )
            i += 1

        current_ea += memory.PTR_SIZE


def vtable_methods_count(cls: tinfo_t, skip_reserved: bool) -> int:
    """Given a class, return the number of entries in it."""
    vtable_ea = vtable_location_from_type(cls)
    if vtable_ea is None:
        return -1

    # Use the iterate_vtable function to count methods if we want to skip reserved methods
    if skip_reserved:
        last_method: VTableItem | None = None
        for method in iterate_vtable(vtable_ea, skip_reserved):
            last_method = method

        if last_method is None:
            return 0

        return last_method.index + 1
    else:
        # If we don't skip reserved methods, we can just count the number of entries in the vtable
        current_ea = vtable_ea + 2 * memory.PTR_SIZE
        i = 0
        while memory.qword_from_ea(current_ea) != 0:
            i += 1
            current_ea += memory.PTR_SIZE
        return i


def vtable_func_at(cls: tinfo_t, offset: int) -> int | None:
    """Given a class and an offset to vtable entry, return the ea at the given offset."""
    vtable_ea = vtable_location_from_type(cls)
    if vtable_ea is None:
        return None

    # Read the func at the relevant offset
    vtable_entry = vtable_ea + (2 * memory.PTR_SIZE + offset)
    vtable_func_ea = memory.qword_from_ea(vtable_entry)
    return vtable_func_ea if functions.is_in_function(vtable_func_ea) else None



__all__ = ["find_func_containing_string", "find_static_caller_for_string", "func_xrefs_to", "get_xrefs_to"]

from collections.abc import Iterator

import ida_hexrays
import ida_xref
import idautils
from ida_hexrays import minsn_t, mop_t

from idahelper import functions, strings
from idahelper.microcode import mba
from idahelper.microcode.visitors import extended_microcode_visitor_t


def get_xrefs_to(ea: int, is_data: bool = False) -> list[int]:
    """Get all xrefs to the given EA"""
    return [xref.frm for xref in idautils.XrefsTo(ea, ida_xref.XREF_DATA if is_data else 0)]


def code_xrefs_to(ea: int) -> list[int]:
    """Get all code xrefs to the given EA"""
    return list(idautils.CodeRefsTo(ea, 1))


def func_xrefs_to(func_ea: int) -> set[int]:
    """Get all xrefs to the given EA, grouped by function"""
    xrefs_in_funcs = set()
    for xref_ea in get_xrefs_to(func_ea):
        func_start = functions.get_start_of_function(xref_ea)
        if func_start is not None:
            xrefs_in_funcs.add(func_start)
    return xrefs_in_funcs


def string_xrefs_to(s: str) -> Iterator[int]:
    """Assume there is a single function that contains s, return the EA of call function"""
    # There might be multiple references to the string, so we need to find the one that is inside a function
    for item in strings.find_strs(s):
        yield from get_xrefs_to(item.ea)
    return None


def find_func_containing_string(s: str) -> int | None:
    """Assume there is a single function that contains s, return the EA of call function"""
    # There might be multiple references to the string, so we need to find the one that is inside a function
    for xref_ea in string_xrefs_to(s):
        func_start = functions.get_start_of_function(xref_ea)
        if func_start is not None:
            return func_start
    return None


def find_static_caller_for_string(s: str) -> int | None:
    """Assume there is a single call(..., s, ...) in code, return the EA of call function"""
    # There might be multiple references to the string, so we need to find the one that is a call
    for item in strings.find_strs(s):
        for func_xref in func_xrefs_to(item.ea):
            finder = CallFinderForEa(item.ea)
            func = mba.from_func(func_xref)
            if func is None:
                print("[Error] Could not build mba for func", hex(func_xref))
                continue
            finder.visit_function(func)
            if finder.result is not None:
                return finder.result

    return None


class CallFinderForEa(extended_microcode_visitor_t):
    def __init__(self, ea: int):
        super().__init__()
        self._ea = ea
        self.result: int | None = None

    def _visit_mop(self, op: mop_t) -> int:
        if op.t == ida_hexrays.mop_v and op.g == self._ea:
            parent = self.parents[-1]
            if not isinstance(parent, mop_t) or parent.t != ida_hexrays.mop_a:
                # If the parent is not an argument, we are not interested in this reference
                print(
                    f"[Warning] Found a reference to the EA that is not a var ref, skipping at {self.top_ins.ea:#x} - {self.top_ins.dstr()}"
                )
                return 0

            parent_args = self.parents[-2]
            if not isinstance(parent_args, mop_t) or parent_args.t != ida_hexrays.mop_f:
                # If the parent is not an instruction, we are not interested in this reference
                print(
                    f"[Warning] Found a reference to the EA that is not a function arguments, skipping at {self.top_ins.ea:#x} - {self.top_ins.dstr()}"
                )
                return 0

            parent_call = self.parents[-3]
            if not isinstance(parent_call, minsn_t) or parent_call.opcode != ida_hexrays.m_call:
                # If the parent is not a call instruction, we are not interested in this reference
                print(
                    f"[Warning] Found a reference to the EA that is not a call instruction, skipping: {parent_call.dstr()}"
                )
                return 0

            caller: mop_t = parent_call.l
            if caller.t != ida_hexrays.mop_v:
                print(f"[Warning] The call is not static, skipping: {parent_call.dstr()}")
                return 0

            self.result = caller.g
            return 1

        return 0



__all__ = [
    "analyze_reg_dependencies",
    "decode_instruction",
    "decode_next_instruction",
    "from_func",
    "get_register_name",
    "is_flow_instruction",
]

import functools
from collections.abc import Iterator

import ida_idp
import ida_ua
import idaapi
import idautils
from ida_funcs import func_t
from ida_ua import insn_t


@functools.cache
def _registers() -> list[str]:
    return ida_idp.ph_get_regnames()


def decode_instruction(ea: int) -> insn_t | None:
    """Decode an instruction at the given ea"""
    return idautils.DecodeInstruction(ea)


def decode_next_instruction(insn: insn_t, func: func_t) -> insn_t | None:
    """Decode the next instruction after the given insn"""
    next_ea = insn.ea + insn.size
    if next_ea >= func.end_ea:
        return None

    return decode_instruction(next_ea)


def decode_previous_instruction(insn: insn_t) -> insn_t | None:
    """Decode the previous instruction for the given insn"""
    return idautils.DecodePrecedingInstruction(insn.ea)[0]


def is_flow_instruction(insn: insn_t) -> bool:
    """Given an instruction, is it possible for it to influence the flow and not running the next instruction in memory"""
    feature = insn.get_canon_feature()
    return (feature & (idaapi.CF_CALL | idaapi.CF_STOP | idaapi.CF_JUMP)) != 0


def get_register_name(reg: int) -> str:
    """Given register index, return its name"""
    return _registers()[reg]


def from_func(func: func_t) -> Iterator[insn_t]:
    for ea in idautils.FuncItems(func.start_ea):
        insn = decode_instruction(ea)
        if insn:
            yield insn


OP_READ = 1
OP_WRITE = 2


def _get_operand_access(insn: insn_t, op_index: int) -> int:
    f = insn.get_canon_feature()
    access = 0
    if f & (ida_idp.CF_USE1 << op_index):
        access |= OP_READ
    if f & (ida_idp.CF_CHG1 << op_index):
        access |= OP_WRITE
    return access


def analyze_reg_dependencies(insn: insn_t) -> tuple[set[str], set[str]]:
    """for the instruction, return two sets: read registers and modified registers"""
    read, write = set(), set()

    for i, op in enumerate(insn.ops):  # type: ignore  # noqa: PGH003
        if op.type == ida_ua.o_void:
            continue

        access = _get_operand_access(insn, i)

        reg_name = None
        if op.type in (ida_ua.o_reg, ida_ua.o_phrase, ida_ua.o_displ, ida_ua.o_idpspec0):
            reg_name = _registers()[op.reg]

        if reg_name:
            if access & OP_READ:
                read.add(reg_name)
            if access & OP_WRITE:
                write.add(reg_name)
    return read, write



from collections.abc import Callable
from functools import wraps
from typing import TypeVar

T = TypeVar("T")


def cache_fast(func: Callable[[], T]) -> Callable[[], T]:
    """Decorator to cache the result of a function for faster access."""
    cached_value: T | None = None

    @wraps(func)
    def wrapper():
        nonlocal cached_value
        if cached_value is None:
            cached_value = func()
        return cached_value

    return wrapper



__all__ = ["PacXplorerClient"]

import sys
from collections import namedtuple
from typing import Protocol

from ..._utils import cache_fast
from ..pac_interface import PacClient, VtableXrefTuple

MovkCodeTuple = namedtuple("MovkCodeTuple", ["pac_tuple", "movk_addr", "trace"])


class VtableAnalyzerProtocol(Protocol):
    def codes_from_func_addr(self, ea: int) -> list: ...

    def func_from_pac_tuple(self, pac_tuple: MovkCodeTuple) -> list[VtableXrefTuple]: ...


class MovkAnalyzerProtocol(Protocol):
    def pac_tuple_from_ea(self, ea: int) -> MovkCodeTuple: ...

    def movks_from_pac_codes(self, pac_codes) -> list[tuple]: ...


class PacxplorerPluginProtocol(Protocol):
    vtable_analyzer: VtableAnalyzerProtocol
    movk_analyzer: MovkAnalyzerProtocol
    analysis_done: bool

    def analyze(self, only_cached=False) -> None: ...


PLUGIN_NAME_CACHED = "pacxplorer_plugin"


class PacXplorerClient(PacClient):
    plugin_name = "pacxplorer"

    @staticmethod
    @cache_fast
    def _get_pac_plugin() -> PacxplorerPluginProtocol:
        # Cache it somewhere else, to avoid analyzing every time we reload our plugin
        main_module = sys.modules["__main__"]
        if hasattr(main_module, PLUGIN_NAME_CACHED):
            return getattr(main_module, PLUGIN_NAME_CACHED)

        PacXplorerClient.ensure_pac_plugin_installed()
        # noinspection PyUnresolvedReferences
        import pacxplorer  # pyright: ignore [reportMissingImports]

        plugin: PacxplorerPluginProtocol = pacxplorer.PacxplorerPlugin()
        plugin.analyze(False)
        if not plugin.analysis_done:
            raise AssertionError("PacExplorer plugin analysis not done, please run the analysis first")
        setattr(main_module, PLUGIN_NAME_CACHED, plugin)
        return plugin

    def _pac_candidates_from_movk(self, movk_ea: int) -> list[VtableXrefTuple]:
        pac_plugin = self._get_pac_plugin()
        return pac_plugin.vtable_analyzer.func_from_pac_tuple(pac_plugin.movk_analyzer.pac_tuple_from_ea(movk_ea))

    def pac_xrefs_to_func(self, func_ea: int) -> list[int]:
        pac_plugin = self._get_pac_plugin()
        pac_codes = pac_plugin.vtable_analyzer.codes_from_func_addr(func_ea)
        if pac_codes is None:
            return []
        movks = pac_plugin.movk_analyzer.movks_from_pac_codes(pac_codes)
        return [addr for addr, code in movks]



def is_objc_method(name: str) -> bool:
    """Does the name look like an Obj-C method?"""
    return len(name) > 3 and name[0] in ["-", "+"] and name[1] == "[" and name[-1] == "]"


def is_objc_static_method(name: str) -> bool:
    """Given obj-C method name, check if it is a static method."""
    return name[0] == "+"



import ida_hexrays
from ida_hexrays import cexpr_t, cinsn_t


def from_expr(expr: cexpr_t, ea: int) -> cinsn_t:
    """Create a cinsn_t from a cexpr_t."""
    new_item = cinsn_t()
    new_item.op = ida_hexrays.cit_expr
    new_item.cexpr = expr
    new_item.thisown = False
    new_item.ea = ea
    return new_item



__all__ = ["VariableModification", "perform_lvar_modifications"]

from dataclasses import dataclass

from ida_hexrays import (
    cfunc_t,
    lvar_saved_info_t,
    lvar_t,
    lvar_uservec_t,
    lvars_t,
    modify_user_lvars,
    user_lvar_modifier_t,
)
from ida_hexrays import (
    rename_lvar as ida_rename_lvar,
)
from ida_typeinf import tinfo_t

from idahelper.ast import cfunc


@dataclass
class VariableModification:
    name: str | None = None
    type: tinfo_t | None = None
    comment: str | None = None
    force_name_change: bool = True


class custom_lvars_modifiers_t(user_lvar_modifier_t):
    def __init__(self, modifications: dict[str, VariableModification]):
        super().__init__()
        self._modifications = modifications

    def modify_lvars(self, lvinf: lvar_uservec_t) -> bool:
        if not self._modifications:
            return False
        has_matched = False
        for lvar in lvinf.lvvec:
            lvar: lvar_saved_info_t
            if (modification := self._modifications.get(lvar.name)) is not None:
                has_matched = True
                custom_lvars_modifiers_t.modify_lvar(lvar, modification)

        return has_matched

    @staticmethod
    def modify_lvar(lvar: lvar_t | lvar_saved_info_t, modification: VariableModification):
        """Modify a single local variable."""
        if modification.name is not None:
            lvar.name = modification.name
            if isinstance(lvar, lvar_t):
                lvar.set_user_name()
        if modification.type is not None:
            lvar.type = modification.type
            if isinstance(lvar, lvar_t):
                lvar.set_user_type()
        if modification.comment is not None:
            lvar.cmt = modification.comment


def perform_lvar_modifications_by_ea(entry_ea: int, modifications: dict[str, VariableModification]) -> bool:
    """Perform the modifications on the local variables of the function."""
    if not modifications:
        return False

    lvars: lvars_t = cfunc.from_ea(entry_ea).get_lvars()
    return perform_lvar_modifications(entry_ea, lvars, modifications)


def perform_lvar_modifications(
    entry_ea: int, lvars: lvars_t, modifications: dict[str, VariableModification], temp_fallback: bool = False
) -> bool:
    """Perform the modifications on the local variables."""
    if not modifications:
        return False

    # According to ida documentation:
    # `lvars.lvvec` contains only variables modified from the defaults.
    # To change other variables, you can, for example, first use rename_lvars, so they get added to this list
    for name, modification in modifications.items():
        if not modification.force_name_change and get_by_name(lvars, name).has_user_name:
            # Already has name, so don't change the name
            # It will be in lvvec, so we can just skip it
            modification.name = None
            continue

        rename_res = ida_rename_lvar(entry_ea, name, name)
        if not rename_res:
            if not temp_fallback:
                print(f"{entry_ea:#x}: Failed to rename local variable {name} to itself, it will not be modified")
            else:
                # IDA API does not support setting local variable's name permanently during decompilation
                # Instead, we will just change it temporarily
                custom_lvars_modifiers_t.modify_lvar(get_by_name(lvars, name), modification)

    return modify_user_lvars(entry_ea, custom_lvars_modifiers_t(modifications))


def rename_lvar(func: cfunc_t | int, old_name: str, new_name: str) -> bool:
    """Rename a local variable in the function."""
    entry_ea = func if isinstance(func, int) else func.entry_ea
    return ida_rename_lvar(entry_ea, old_name, new_name)


def get_by_name(lvars: lvars_t, name: str) -> lvar_t:
    """Get the local variable with the given name, raise exception if it does not exist."""
    for lvar in lvars:
        if lvar.name == name:
            return lvar

    raise ValueError(f"Local variable {name} not found in {lvars}")


def get_index_by_name(lvars: lvars_t, name: str) -> int:
    """Get the index of the local variable with the given name."""
    for i, lvar in enumerate(lvars):
        if lvar.name == name:
            return i
    return -1


def get_index(lvars: lvars_t, lvar: lvar_t) -> int:
    """Get the index of the local variable with the given name."""
    for i, lvar2 in enumerate(lvars):
        if lvar == lvar2:
            return i
    return -1


def get_flags(lvar: lvar_t) -> int:  # noqa: C901
    """Given lvar, return the int value of the CVAR_ flags"""
    flags = 0

    if lvar.used:
        flags |= 0x00000001  # CVAR_USED
    if lvar.typed:
        flags |= 0x00000002  # CVAR_TYPE
    if lvar.has_nice_name:
        flags |= 0x00000004  # CVAR_NAME
    if lvar.mreg_done:
        flags |= 0x00000008  # CVAR_MREG
    if lvar.is_unknown_width:
        flags |= 0x00000010  # CVAR_NOWD
    if lvar.has_user_name:
        flags |= 0x00000020  # CVAR_UNAME
    if lvar.has_user_type:
        flags |= 0x00000040  # CVAR_UTYPE
    if lvar.is_result_var:
        flags |= 0x00000080  # CVAR_RESULT
    if lvar.is_arg_var:
        flags |= 0x00000100  # CVAR_ARG
    if lvar.is_fake_var:
        flags |= 0x00000200  # CVAR_FAKE
    if lvar.is_overlapped_var:
        flags |= 0x00000400  # CVAR_OVER
    if lvar.is_floating_var:
        flags |= 0x00000800  # CVAR_FLOAT
    if lvar.is_spoiled_var:
        flags |= 0x00001000  # CVAR_SPOILED
    if lvar.is_mapdst_var:
        flags |= 0x00002000  # CVAR_MAPDST
    if lvar.is_thisarg():
        flags |= 0x00008000  # CVAR_THISARG
    if lvar.is_split_var():
        flags |= 0x00010000  # CVAR_SPLIT
    if lvar.has_regname():
        flags |= 0x00020000  # CVAR_REGNAME
    if lvar.is_noptr_var():
        flags |= 0x00040000  # CVAR_NOPTR
    if lvar.is_dummy_arg():
        flags |= 0x00080000  # CVAR_DUMMY
    if lvar.is_notarg():
        flags |= 0x00100000  # CVAR_NOTARG
    if lvar.is_automapped():
        flags |= 0x00200000  # CVAR_AUTOMAP
    if lvar.is_used_byref():
        flags |= 0x00400000  # CVAR_BYREF
    if lvar.in_asm():
        flags |= 0x00800000  # CVAR_INASM
    if lvar.is_decl_unused():
        flags |= 0x01000000  # CVAR_UNUSED
    if lvar.is_shared():
        flags |= 0x02000000  # CVAR_SHARED
    return flags
